@@ -3,8 +3,9 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2000-2020, University of Amsterdam
-                              Vu University Amsterdam
+    Copyright (c)  2000-2021, University of Amsterdam
+                              VU University Amsterdam
+			      SWI-Prolog Solutions b.v.
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -39,7 +40,7 @@
 #include <SWI-Prolog.h>
 #include <string.h>
 #include <wchar.h>
-#ifndef __APPLE__
+#if !(defined(__APPLE__) || defined(__FreeBSD__))
 #include <malloc.h>
 #endif
 
@@ -79,6 +80,14 @@ public:
   PlFunctor(const wchar_t *name, ARITY_T arity)
   { functor = PL_new_functor(PL_new_atom_wchars(wcslen(name), name), arity);
   }
+
+  operator const functor_t(void) const
+  { return functor;
+  }
+
+  int operator ==(functor_t to) const
+  { return functor == to;
+  }
 };
 
 
@@ -130,6 +139,7 @@ public:
   term_t ref;
 
   PlTerm();
+  PlTerm(const PlTerm &other) : ref(other.ref) {}
   PlTerm(term_t t)
   { ref = t;
   }
@@ -215,13 +225,21 @@ class PlTermv
 {
 public:
   term_t a0;
-  int    size;
+  size_t size;
 
   PlTermv(int n)
   { a0   = PL_new_term_refs(n);
-    size = n;
+    size = static_cast<size_t>(n);
   }
   PlTermv(int n, term_t t0)
+  { a0   = t0;
+    size = static_cast<size_t>(n);
+  }
+  PlTermv(size_t n)
+  { a0   = PL_new_term_refs(static_cast<int>(n));
+    size = n;
+  }
+  PlTermv(size_t n, term_t t0)
   { a0   = t0;
     size = n;
   }
@@ -233,7 +251,7 @@ public:
   PlTermv(PlTerm m0, PlTerm m1, PlTerm m2, PlTerm m3);
   PlTermv(PlTerm m0, PlTerm m1, PlTerm m2, PlTerm m3, PlTerm m4);
 
-  PlTerm operator [](int n) const;
+  PlTerm operator [](size_t n) const;
 };
 
 		 /*******************************
@@ -302,8 +320,8 @@ public:
   operator const char *(void);
   operator const wchar_t *(void);
 
-  int plThrow()
-  { return PL_raise_exception(ref);
+  foreign_t plThrow()
+  { return static_cast<foreign_t>(PL_raise_exception(ref));
   }
 
   void cppThrow();
@@ -416,12 +434,12 @@ class PlTermvDomainError : public PlException
 {
 public:
 
-  PlTermvDomainError(int size, int n) :
+  PlTermvDomainError(size_t size, size_t n) :
     PlException(PlCompound("error",
 			   PlTermv(PlCompound("domain_error",
 					      PlTermv(PlCompound("argv",
 								 size),
-						      PlTerm((long)n))),
+						      PlTerm(static_cast<long>(n)))),
 				   PlTerm())))
   {
   }
@@ -448,7 +466,7 @@ PlTerm::PlTerm(const char *text)
 __inline
 PlTerm::PlTerm(const wchar_t *text)
 { if ( !(ref = PL_new_term_ref()) ||
-       !PL_unify_wchars(ref, PL_ATOM, (size_t)-1, text) )
+       !PL_unify_wchars(ref, PL_ATOM, static_cast<size_t>(-1), text) )
     throw PlResourceError();
 }
 
@@ -461,9 +479,8 @@ PlTerm::PlTerm(long val)
 
 __inline
 PlTerm::PlTerm(double val)
-{ ref = PL_new_term_ref();
-
-  if ( !PL_put_float(ref, val) )
+{ if ( !(ref = PL_new_term_ref()) ||
+       !PL_put_float(ref, val) )
     throw PlResourceError();
 }
 
@@ -500,7 +517,7 @@ PlString::PlString(const char *text, size_t len) : PlTerm()
 
 __inline
 PlString::PlString(const wchar_t *text) : PlTerm()
-{ if ( !PL_unify_wchars(ref, PL_STRING, (size_t)-1, text) )
+{ if ( !PL_unify_wchars(ref, PL_STRING, static_cast<size_t>(-1), text) )
     throw PlResourceError();
 }
 
@@ -524,13 +541,13 @@ PlCharList::PlCharList(const char *text) : PlTerm()
 
 __inline
 PlCodeList::PlCodeList(const wchar_t *text) : PlTerm()
-{ if ( !PL_unify_wchars(ref, PL_CODE_LIST, (size_t)-1, text) )
+{ if ( !PL_unify_wchars(ref, PL_CODE_LIST, static_cast<size_t>(-1), text) )
     throw PlResourceError();
 }
 
 __inline
 PlCharList::PlCharList(const wchar_t *text) : PlTerm()
-{ if ( !PL_unify_wchars(ref, PL_CHAR_LIST, (size_t)-1, text) )
+{ if ( !PL_unify_wchars(ref, PL_CHAR_LIST, static_cast<size_t>(-1), text) )
     throw PlResourceError();
 }
 
@@ -645,14 +662,14 @@ public:
   qid_t qid;
 
   PlQuery(predicate_t pred, const PlTermv &av)
-  { qid = PL_open_query((module_t)0, PL_Q_PASS_EXCEPTION, pred, av.a0);
+  { qid = PL_open_query(static_cast<module_t>(0), PL_Q_PASS_EXCEPTION, pred, av.a0);
     if ( !qid )
       throw PlResourceError();
   }
   PlQuery(const char *name, const PlTermv &av)
-  { predicate_t p = PL_predicate(name, av.size, "user");
+  { predicate_t p = PL_predicate(name, static_cast<int>(av.size), "user");
 
-    qid = PL_open_query((module_t)0, PL_Q_PASS_EXCEPTION, p, av.a0);
+    qid = PL_open_query(static_cast<module_t>(0), PL_Q_PASS_EXCEPTION, p, av.a0);
     if ( !qid )
       throw PlResourceError();
   }
@@ -799,9 +816,9 @@ PlTerm::operator [](ARITY_T index) const
     return t;
 
   if ( !PL_is_compound(ref) )
-    throw PlTypeError("compound", ref);
-  else
-  { if ( !PL_put_integer(t.ref, index) )
+  { throw PlTypeError("compound", ref);
+  } else
+  { if ( !PL_put_uint64(t.ref, index) )
       throw PlResourceError();
 
     if ( index < 1 )
@@ -866,7 +883,7 @@ __inline int PlTerm::operator =(const char *v)		/* term = atom */
 }
 
 __inline int PlTerm::operator =(const wchar_t *v)	/* term = atom */
-{ int rc = PL_unify_wchars(ref, PL_ATOM, (size_t)-1, v);
+{ int rc = PL_unify_wchars(ref, PL_ATOM, static_cast<size_t>(-1), v);
   term_t ex;
 
   if ( !rc && (ex=PL_exception(0)) )
@@ -1001,6 +1018,12 @@ __inline int PlTerm::operator ==(const PlAtom &a) const
 		 *	   COMPOUND (BODY)	*
 		 *******************************/
 
+__inline void
+PlPutTerm(term_t to, term_t from)
+{ if ( !PL_put_term(to, from) )
+    throw PlResourceError();
+}
+
 
 __inline
 PlCompound::PlCompound(const char *text) : PlTerm()
@@ -1009,7 +1032,7 @@ PlCompound::PlCompound(const char *text) : PlTerm()
   if ( !PL_chars_to_term(text, t) )
     throw PlException(t);
 
-  PL_put_term(ref, t);
+  PlPutTerm(ref, t);
 }
 
 __inline
@@ -1019,7 +1042,7 @@ PlCompound::PlCompound(const wchar_t *text) : PlTerm()
   if ( !PL_wchars_to_term(text, t) )
     throw PlException(t);
 
-  PL_put_term(ref, t);
+  PlPutTerm(ref, t);
 }
 
 __inline
@@ -1054,28 +1077,28 @@ __inline PlTermv::PlTermv(PlTerm m0, PlTerm m1)
 { size = 2;
   if ( !(a0 = PL_new_term_refs(2)) )
     throw PlResourceError();
-  PL_put_term(a0+0, m0);
-  PL_put_term(a0+1, m1);
+  PlPutTerm(a0+0, m0);
+  PlPutTerm(a0+1, m1);
 }
 
 __inline PlTermv::PlTermv(PlTerm m0, PlTerm m1, PlTerm m2)
 { size = 3;
   if ( !(a0 = PL_new_term_refs(3)) )
     throw PlResourceError();
-  PL_put_term(a0+0, m0);
-  PL_put_term(a0+1, m1);
-  PL_put_term(a0+2, m2);
+  PlPutTerm(a0+0, m0);
+  PlPutTerm(a0+1, m1);
+  PlPutTerm(a0+2, m2);
 }
 
 __inline PlTermv::PlTermv(PlTerm m0, PlTerm m1, PlTerm m2, PlTerm m3)
 { size = 4;
   if ( !(a0 = PL_new_term_refs(4)) )
     throw PlResourceError();
-  PL_put_term(a0+0, m0);
-  PL_put_term(a0+1, m1);
-  PL_put_term(a0+2, m2);
+  PlPutTerm(a0+0, m0);
+  PlPutTerm(a0+1, m1);
+  PlPutTerm(a0+2, m2);
 
-  PL_put_term(a0+3, m3);
+  PlPutTerm(a0+3, m3);
 }
 
 __inline PlTermv::PlTermv(PlTerm m0, PlTerm m1, PlTerm m2,
@@ -1083,17 +1106,17 @@ __inline PlTermv::PlTermv(PlTerm m0, PlTerm m1, PlTerm m2,
 { size = 5;
   if ( !(a0 = PL_new_term_refs(5)) )
     throw PlResourceError();
-  PL_put_term(a0+0, m0);
-  PL_put_term(a0+1, m1);
-  PL_put_term(a0+2, m2);
-  PL_put_term(a0+3, m3);
-  PL_put_term(a0+4, m4);
+  PlPutTerm(a0+0, m0);
+  PlPutTerm(a0+1, m1);
+  PlPutTerm(a0+2, m2);
+  PlPutTerm(a0+3, m3);
+  PlPutTerm(a0+4, m4);
 }
 
 
 __inline PlTerm
-PlTermv::operator [](int n) const
-{ if ( n < 0 || n >= size )
+PlTermv::operator [](size_t n) const
+{ if ( n >= size )
     throw PlTermvDomainError(size, n);
 
   return PlTerm(a0+n);
@@ -1119,7 +1142,7 @@ __inline PlException::operator const char *(void)
   av[0] = PlTerm(ref);
   PlQuery q("$messages", "message_to_string", av);
   if ( q.next_solution() )
-    return (char *)av[1];
+    return static_cast<char*>(av[1]);
 #endif
   return "[ERROR: Failed to generate message.  Internal error]\n";
 }
@@ -1140,7 +1163,7 @@ __inline PlException::operator const wchar_t *(void)
   av[0] = PlTerm(ref);
   PlQuery q("$messages", "message_to_string", av);
   if ( q.next_solution() )
-    return (wchar_t *)av[1];
+    return static_cast<wchar_t*>(av[1]);
 #endif
   return L"[ERROR: Failed to generate message.  Internal error]\n";
 }
@@ -1229,7 +1252,7 @@ public:
 
   PlEngine(char *av0)
   { int ac = 0;
-    char **av = (char **)malloc(sizeof(char *) * 2);
+    char **av = static_cast<char**>(malloc(sizeof(char *) * 2));
 
     av[ac++] = av0;
 
