@@ -48,7 +48,6 @@
 #include "config/win32.h"
 #endif
 
-#define PL_ARITY_AS_SIZE 1
 #include <tchar.h>
 #include <malloc.h>
 #include <stdio.h>
@@ -248,22 +247,23 @@ Srlc_write(void *handle, char *buffer, size_t size)
 }
 
 
+
 static int
 Srlc_close(void *handle)
 { rlc_console c = handle;
   uintptr_t v;
   int closed = 0;
 
-  if ( rlc_get(handle, RLC_PROLOG_INPUT, &v) && v &&
-       ((IOSTREAM *)v)->flags && SIO_CLOSING )
+#define ison(p,f) (((p)->flags)&(f))
+#define IS_CLOSING(v) ison((IOSTREAM *)v, SIO_CLOSING)
+
+  if ( rlc_get(handle, RLC_PROLOG_INPUT, &v) && v && IS_CLOSING(v) )
   { rlc_set(handle, RLC_PROLOG_INPUT, 0L, NULL);
     closed++;
-  } else if ( rlc_get(handle, RLC_PROLOG_OUTPUT, &v) && v &&
-	      ((IOSTREAM *)v)->flags && SIO_CLOSING )
+  } else if ( rlc_get(handle, RLC_PROLOG_OUTPUT, &v) && v && IS_CLOSING(v) )
   { rlc_set(handle, RLC_PROLOG_OUTPUT, 0L, NULL);
     closed++;
-  } else if ( rlc_get(handle, RLC_PROLOG_ERROR, &v) && v &&
-	      ((IOSTREAM *)v)->flags && SIO_CLOSING )
+  } else if ( rlc_get(handle, RLC_PROLOG_ERROR, &v) && v && IS_CLOSING(v) )
   { rlc_set(handle, RLC_PROLOG_ERROR, 0L, NULL);
   }
 
@@ -365,7 +365,7 @@ pl_win_open_console(term_t title, term_t input, term_t output, term_t error,
   size_t len;
 
   memset(&attr, 0, sizeof(attr));
-  if ( !PL_get_wchars(title, &len, &s, CVT_ALL|BUF_RING|CVT_EXCEPTION) )
+  if ( !PL_get_wchars(title, &len, &s, CVT_ALL|BUF_STACK|CVT_EXCEPTION) )
     return FALSE;
   attr.title = (const TCHAR*) s;
 
@@ -595,7 +595,7 @@ get_chars_arg_ex(int a, term_t t, TCHAR **v)
 { term_t arg = PL_new_term_ref();
 
   if ( PL_get_arg(a, t, arg) &&
-       PL_get_wchars(arg, NULL, v, CVT_ALL|BUF_RING|CVT_EXCEPTION) )
+       PL_get_wchars(arg, NULL, v, CVT_ALL|BUF_STACK|CVT_EXCEPTION) )
     return TRUE;
 
   return FALSE;
