@@ -3,7 +3,7 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  1985-2021, University of Amsterdam
+    Copyright (c)  1985-2022, University of Amsterdam
                               VU University Amsterdam
 			      CWI, Amsterdam
 			      SWI-Prolog Solutions b.v.
@@ -64,6 +64,7 @@
 #include "pl-supervisor.h"
 #include "pl-index.h"
 #include "pl-cont.h"
+#include "pl-coverage.h"
 #include <fenv.h>
 #ifdef _MSC_VER
 #pragma warning(disable: 4102)		/* unreferenced labels */
@@ -273,7 +274,14 @@ void
 updateAlerted(PL_local_data_t *ld)
 { int mask = 0;
 
-  WITH_LD(ld) if ( is_signalled() )		mask |= ALERT_SIGNAL;
+  WITH_LD(ld)
+  { if ( is_signalled() )		        mask |= ALERT_SIGNAL;
+    if ( !truePrologFlag(PLFLAG_VMI_BUILTIN) ||
+	 ld->prolog_flag.occurs_check != OCCURS_CHECK_FALSE )
+      ld->slow_unify = TRUE;  /* see VMI B_UNIFY_VAR */
+    else
+      ld->slow_unify = FALSE;
+  }
 #ifdef O_PROFILE
   if ( ld->profile.active )			mask |= ALERT_PROFILE;
 #endif
@@ -298,13 +306,9 @@ updateAlerted(PL_local_data_t *ld)
 #endif
   if ( ld->fli.string_buffers.top )		mask |= ALERT_BUFFER;
   if ( UNDO_SCHEDULED(ld) )			mask |= ALERT_UNDO;
+  if ( LD->coverage.active )			mask |= ALERT_COVERAGE;
 
   ld->alerted = mask;
-
-  if ( (mask&ALERT_DEBUG) || ld->prolog_flag.occurs_check != OCCURS_CHECK_FALSE )
-    ld->slow_unify = TRUE;		/* see VMI B_UNIFY_VAR */
-  else
-    ld->slow_unify = FALSE;
 }
 
 
@@ -741,14 +745,13 @@ As frameFinished() is rather time critical this seems worthwhile.
 #define call1(mdef, goal) LDFUNC(call1, mdef, goal)
 static int
 call1(DECL_LD Module mdef, term_t goal)
-{ static predicate_t PRED_call1 = NULL;
+{ predicate_t pred;
   qid_t qid;
   int rc;
 
-  if ( !PRED_call1 )
-    PRED_call1 = PL_predicate("call", 1, "system");
+  pred = _PL_predicate("call", 1, "system", &GD->procedures.call1);
 
-  qid = PL_open_query(mdef, PL_Q_PASS_EXCEPTION, PRED_call1, goal);
+  qid = PL_open_query(mdef, PL_Q_PASS_EXCEPTION, pred, goal);
   rc = PL_next_solution(qid);
   PL_cut_query(qid);
 
@@ -844,9 +847,9 @@ callCleanupHandler(DECL_LD LocalFrame fr, enum finished reason)
       if ( saveWakeup(&wstate, FALSE) )
       { int rval;
 
-	startCritical;
+	startCritical();
         rval = call_term(contextModule(fr), clean);
-	if ( !endCritical )
+	if ( !endCritical() )
 	  rval = FALSE;
 	if ( !rval && exception_term )
 	  wstate.flags |= WAKEUP_KEEP_URGENT_EXCEPTION;
@@ -1263,9 +1266,9 @@ put_vm_call(DECL_LD term_t t, term_t frref, Code PC, code op, int has_firstvar,
 
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-callBreakHook() calls prolog:break_hook/6 as
+callBreakHook() calls prolog:break_hook/7 as
 
-    prolog:break_hook(+Clause, +PC, +Frame, +Choice, +Goal, -Action) is semidet.
+    prolog:break_hook(+Clause, +PC, +Frame, +Choice, +Goal, +Debug, -Action) is semidet.
 
 (*) If put_vm_call() addresses  `F`  (first   var)  variables,  it  will
 initialise these to bind to the  goal.   However,  if GC comes along, it
@@ -1291,8 +1294,8 @@ callBreakHook(DECL_LD LocalFrame frame, Choice bfr,
   { LD->slow_unify = TRUE;
     goto default_action;
   }
-  proc = _PL_predicate("break_hook", 6, "prolog",
-		       &GD->procedures.prolog_break_hook6);
+  proc = _PL_predicate("break_hook", 7, "prolog",
+		       &GD->procedures.prolog_break_hook7);
   if ( !getProcDefinition(proc)->impl.any.defined )
     goto default_action;
 
@@ -1315,7 +1318,7 @@ callBreakHook(DECL_LD LocalFrame frame, Choice bfr,
 
   if ( saveWakeup(&wstate, FALSE) )
   { if ( (cid=PL_open_foreign_frame()) )
-    { term_t argv = PL_new_term_refs(6);
+    { term_t argv = PL_new_term_refs(7);
       Clause clause = frame->clause->value.clause;
       qid_t qid;
 
@@ -1324,6 +1327,7 @@ callBreakHook(DECL_LD LocalFrame frame, Choice bfr,
       PL_put_intptr(argv+1, PC - clause->codes);
       PL_put_frame(argv+2, frame);
       PL_put_choice(argv+3, bfr);
+      PL_put_bool(argv+5, debugstatus.debugging);
       if ( ( op == B_UNIFY_EXIT &&
              put_call_goal(argv+4, GD->procedures.equals2) &&
              PL_cons_functor_v(argv+4, FUNCTOR_call1, argv+4) ) ||
@@ -1342,7 +1346,7 @@ callBreakHook(DECL_LD LocalFrame frame, Choice bfr,
 	  { atom_t a_action;
 	    break_action action;
 
-	    if ( PL_get_atom(argv+5, &a_action) )
+	    if ( PL_get_atom(argv+6, &a_action) )
 	    { if ( a_action == ATOM_continue )
 	      { action = BRK_CONTINUE;
 	      } else if ( a_action == ATOM_trace )
@@ -1356,9 +1360,9 @@ callBreakHook(DECL_LD LocalFrame frame, Choice bfr,
 	      restoreWakeup(&wstate);
 
 	      return action;
-	    } else if ( PL_is_functor(argv+5, FUNCTOR_call1) )
+	    } else if ( PL_is_functor(argv+6, FUNCTOR_call1) )
 	    { LocalFrame NFR = LD->query->next_environment;
-	      Word p = valTermRef(argv+5);
+	      Word p = valTermRef(argv+6);
 
 	      deRef(p);
 	      assert(hasFunctor(p[0], FUNCTOR_call1));
@@ -1372,7 +1376,7 @@ callBreakHook(DECL_LD LocalFrame frame, Choice bfr,
 	      return BRK_CALL;
 	    } else
 	    { invalid_action:
-	      PL_warning("prolog:break_hook/6: invalid action");
+	      PL_warning("prolog:break_hook/7: invalid action");
 	    }
 	  }
 	}
@@ -1544,6 +1548,10 @@ destroyLocalDefinition(Definition def, unsigned int tid)
   LocalDefinitions v = def->impl.local.local;
   Definition local;
 
+  DEBUG(MSG_PRED_COUNT,
+	Sdprintf("Free local def[%d] for %s at %p\n",
+		 tid, predicateName(def), def));
+
   local = v->blocks[idx][tid];
   v->blocks[idx][tid] = NULL;
   destroyDefinition(local);
@@ -2144,11 +2152,11 @@ exception_hook(DECL_LD qid_t pqid, term_t fr, term_t catchfr_ref)
       { PL_put_frame(av+3, NULL);	/* puts 'none' */
       }
 
-      startCritical;
+      startCritical();
       qid = PL_open_query(MODULE_user, PL_Q_NODEBUG|PL_Q_CATCH_EXCEPTION,
 			  PROCEDURE_exception_hook4, av);
       rc = PL_next_solution(qid);
-      rc = endCritical && rc;
+      rc = endCritical() && rc;
       debug = debugstatus.debugging;
       trace = debugstatus.tracing;
       if ( rc )				/* pass user setting trace/debug */
@@ -2549,7 +2557,7 @@ one has a programPointer pointing to  I_EXITQUERY, doing the return from
 a query.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-#define NDEBUG_SAVE_FLAGS prologFlagMask(PLFLAG_LASTCALL)
+#define NDEBUG_SAVE_FLAGS RUN_MODE_NORMAL
 
 qid_t
 PL_open_query(Module ctx, int flags, Procedure proc, term_t args)
@@ -2679,7 +2687,7 @@ PL_open_query(Module ctx, int flags, Procedure proc, term_t args)
     qf->debugSave = debugstatus.debugging;
     debugstatus.debugging = DBG_OFF;
     qf->flags_saved = (LD->prolog_flag.mask.flags[0] & NDEBUG_SAVE_FLAGS);
-    setPrologFlagMask(PLFLAG_LASTCALL);
+    setPrologRunMode(RUN_MODE_NORMAL);
 #ifdef O_LIMIT_DEPTH
     qf->saved_depth_limit   = LD->depth_info.limit;
     qf->saved_depth_reached = LD->depth_info.reached;
